# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for testing Reverb adders."""

from typing import Any, Callable, Optional, Sequence, Tuple, TypeVar, Union

from absl.testing import absltest
from acme import specs
from acme import types
from acme.adders import base as adders_base
from acme.adders import reverb as adders
from acme.utils import tree_utils
import dm_env
import numpy as np
import reverb
import tensorflow as tf
import tree

StepWithExtra = Tuple[Any, dm_env.TimeStep, Any]
StepWithoutExtra = Tuple[Any, dm_env.TimeStep]
Step = TypeVar('Step', StepWithExtra, StepWithoutExtra)


def make_trajectory(observations):
  """Make a simple trajectory from a sequence of observations.

  Arguments:
    observations: a sequence of observations.

  Returns:
    a tuple (first, steps) where first contains the initial dm_env.TimeStep
    object and steps contains a list of (action, step) tuples. The length of
    steps is given by episode_length.
  """
  first = dm_env.restart(observations[0])
  middle = [(0, dm_env.transition(reward=0.0, observation=observation))
            for observation in observations[1:-1]]
  last = (0, dm_env.termination(reward=0.0, observation=observations[-1]))
  return first, middle + [last]


def make_sequence(observations):
  """Create a sequence of timesteps of the form `first, [second, ..., last]`."""
  first, steps = make_trajectory(observations)
  observation = first.observation
  sequence = []
  start_of_episode = True
  for action, timestep in steps:
    extras = ()
    sequence.append((observation, action, timestep.reward, timestep.discount,
                     start_of_episode, extras))
    observation = timestep.observation
    start_of_episode = False
  sequence.append((observation, 0, 0.0, 0.0, False, ()))
  return sequence


def _numeric_to_spec(x: Union[float, int, np.ndarray]):
  if isinstance(x, np.ndarray):
    return specs.Array(shape=x.shape, dtype=x.dtype)
  elif isinstance(x, (float, int)):
    return specs.Array(shape=(), dtype=type(x))
  else:
    raise ValueError(f'Unsupported numeric: {type(x)}')


def get_specs(step):
  """Infer spec from an example step."""
  env_spec = tree.map_structure(
      _numeric_to_spec,
      specs.EnvironmentSpec(
          observations=step[1].observation,
          actions=step[0],
          rewards=step[1].reward,
          discounts=step[1].discount))

  has_extras = len(step) == 3
  if has_extras:
    extras_spec = tree.map_structure(_numeric_to_spec, step[2])
  else:
    extras_spec = ()

  return env_spec, extras_spec


class AdderTestMixin(absltest.TestCase):
  """A helper mixin for testing Reverb adders.

  Note that any test inheriting from this mixin must also inherit from something
  that provides the Python unittest assert methods.
  """

  server: reverb.Server
  client: reverb.Client

  @classmethod
  def setUpClass(cls):
    super().setUpClass()

    replay_table = reverb.Table.queue(adders.DEFAULT_PRIORITY_TABLE, 1000)
    cls.server = reverb.Server([replay_table])
    cls.client = reverb.Client(f'localhost:{cls.server.port}')

  def tearDown(self):
    self.client.reset(adders.DEFAULT_PRIORITY_TABLE)
    super().tearDown()

  @classmethod
  def tearDownClass(cls):
    cls.server.stop()
    super().tearDownClass()

  def num_episodes(self):
    info = self.client.server_info(1)[adders.DEFAULT_PRIORITY_TABLE]
    return info.num_episodes

  def num_items(self):
    info = self.client.server_info(1)[adders.DEFAULT_PRIORITY_TABLE]
    return info.current_size

  def items(self):
    sampler = self.client.sample(
        table=adders.DEFAULT_PRIORITY_TABLE,
        num_samples=self.num_items(),
        emit_timesteps=False)
    return [sample.data for sample in sampler]  # pytype: disable=attribute-error

  def run_test_adder(
      self,
      adder: adders_base.Adder,
      first: dm_env.TimeStep,
      steps: Sequence[Step],
      expected_items: Sequence[Any],
      signature: types.NestedSpec,
      pack_expected_items: bool = False,
      stack_sequence_fields: bool = True,
      repeat_episode_times: int = 1,
      end_behavior: adders.EndBehavior = adders.EndBehavior.ZERO_PAD,
      item_transform: Optional[Callable[[Sequence[np.ndarray]], Any]] = None):
    """Runs a unit test case for the adder.

    Args:
      adder: The instance of `Adder` that is being tested.
      first: The first `dm_env.TimeStep` that is used to call
        `Adder.add_first()`.
      steps: A sequence of (action, timestep) tuples that are passed to
        `Adder.add()`.
      expected_items: The sequence of items that are expected to be created
        by calling the adder's `add_first()` method on `first` and `add()` on
        all of the elements in `steps`.
      signature: Signature that written items must be compatible with.
      pack_expected_items: Deprecated and not used. If true the expected items
        are given unpacked and need to be packed in a list before comparison.
      stack_sequence_fields: Whether to stack the sequence fields of the
        expected items before comparing to the observed items. Usually False
        for transition adders and True for both episode and sequence adders.
      repeat_episode_times: How many times to run an episode.
      end_behavior: How end of episode should be handled.
      item_transform: Transformation of item simulating the work done by the
        dataset pipeline on the learner in a real setup.
    """

    del pack_expected_items

    if not steps:
      raise ValueError('At least one step must be given.')

    has_extras = len(steps[0]) == 3
    for episode_id in range(repeat_episode_times):
      # Add all the data up to the final step.
      adder.add_first(first)
      for step in steps[:-1]:
        action, ts = step[0], step[1]

        if has_extras:
          extras = step[2]
        else:
          extras = ()

        adder.add(action, next_timestep=ts, extras=extras)

      # Add the final step.
      adder.add(*steps[-1])

    # Force run the destructor to trigger the flushing of all pending items.
    getattr(adder, '__del__', lambda: None)()

    # Ending the episode should close the writer. No new writer should yet have
    # been created as it is constructed lazily.
    if end_behavior is not adders.EndBehavior.CONTINUE:
      self.assertEqual(self.num_episodes(), repeat_episode_times)

    # Make sure our expected and observed data match.
    observed_items = self.items()

    # Check matching number of items.
    self.assertEqual(len(expected_items), len(observed_items))

    # Check items are matching according to numpy's almost_equal.
    for expected_item, observed_item in zip(expected_items, observed_items):
      if stack_sequence_fields:
        expected_item = tree_utils.stack_sequence_fields(expected_item)

      # Apply the transformation which would be done by the dataset in a real
      # setup.
      if item_transform:
        observed_item = item_transform(observed_item)

      tree.map_structure(np.testing.assert_array_almost_equal,
                         tree.flatten(expected_item),
                         tree.flatten(observed_item))

    # Make sure the signature matches was is being written by Reverb.
    def _check_signature(spec: tf.TensorSpec, value: np.ndarray):
      self.assertTrue(spec.is_compatible_with(tf.convert_to_tensor(value)))

    # Check that it is possible to unpack observed using the signature.
    for item in observed_items:
      tree.map_structure(_check_signature, tree.flatten(signature),
                         tree.flatten(item))



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the D4PG agent class, using JAX."""

import functools
from typing import Callable, Optional

from acme import specs
from acme.agents.jax.d4pg import builder
from acme.jax import utils
from acme.jax.layouts import distributed_layout
from acme.jax.layouts import local_layout
from acme.utils import counting
from acme.utils import loggers
import dm_env


NetworkFactory = Callable[[specs.EnvironmentSpec], builder.D4PGNetworks]


class DistributedD4PG(distributed_layout.DistributedLayout):
  """Program definition for D4PG.

  We are in the process of migrating towards a more modular agent configuration.
  This is maintained now for compatibility.
  """

  def __init__(
      self,
      environment_factory: Callable[[bool], dm_env.Environment],
      network_factory: NetworkFactory,
      random_seed: int,
      num_actors: int,
      environment_spec: Optional[specs.EnvironmentSpec] = None,
      batch_size: int = 256,
      prefetch_size: int = 2,
      min_replay_size: int = 1000,
      max_replay_size: int = 1000000,
      samples_per_insert: float = 32.0,
      n_step: int = 5,
      sigma: float = 0.3,
      clipping: bool = True,
      discount: float = 0.99,
      target_update_period: int = 100,
      device_prefetch: bool = True,
      log_to_bigtable: bool = False,
      log_every: float = 10.0,
  ):
    config = builder.D4PGConfig(
        discount=discount,
        learning_rate=1e-4,
        batch_size=batch_size,
        prefetch_size=prefetch_size,
        target_update_period=target_update_period,
        min_replay_size=min_replay_size,
        max_replay_size=max_replay_size,
        samples_per_insert=samples_per_insert,
        n_step=n_step,
        sigma=sigma,
        clipping=clipping,
    )
    logger_fn = functools.partial(
        loggers.make_default_logger,
        'learner',
        log_to_bigtable,
        time_delta=log_every,
        asynchronous=True,
        serialize_fn=utils.fetch_devicearray,
        steps_key='learner_steps')
    d4pg_builder = builder.D4PGBuilder(config, logger_fn=logger_fn)

    super().__init__(
        seed=random_seed,
        environment_factory=lambda: environment_factory(False),
        network_factory=network_factory,
        builder=d4pg_builder,
        policy_network=lambda n: builder.get_default_behavior_policy(n, config),
        evaluator_factories=[
            distributed_layout.default_evaluator(
                environment_factory=lambda: environment_factory(True),
                network_factory=network_factory,
                builder=d4pg_builder,
                policy_factory=builder.get_default_eval_policy,
                log_to_bigtable=log_to_bigtable)
        ],
        num_actors=num_actors,
        environment_spec=environment_spec,
        device_prefetch=device_prefetch,
        log_to_bigtable=log_to_bigtable,
        actor_logger_fn=distributed_layout.get_default_logger_fn(
            log_to_bigtable, log_every),
        prefetch_size=config.prefetch_size,
    )


class D4PG(local_layout.LocalLayout):
  """Local agent for D4PG.
  """

  def __init__(
      self,
      spec: specs.EnvironmentSpec,
      network: builder.D4PGNetworks,
      config: builder.D4PGConfig,
      random_seed: int,
      counter: Optional[counting.Counter] = None,
  ):
    # In the case of a synchronous agent, we do not use Reverb's built-in rate
    # limitation to avoid deadlocks; so rather than using the Builder's min
    # replay size, we trivially set it to 1 after extracting the requested value
    # to be honored by the synchronous Layout instead.
    min_replay_size = config.min_replay_size
    config.min_replay_size = 1

    # Local layout (actually agent.Agent) makes sure that we populate the
    # buffer with min_replay_size initial transitions and that there's no need
    # for tolerance_rate. In order to avoid deadlocks we disable rate limiting.
    # This is achieved by setting the rate tolerance to be infinite.
    config.samples_per_insert_tolerance_rate = float('inf')

    self.builder = builder.D4PGBuilder(config)
    super().__init__(
        seed=random_seed,
        environment_spec=spec,
        builder=self.builder,
        networks=network,
        policy_network=builder.get_default_behavior_policy(network, config),
        batch_size=config.batch_size,
        samples_per_insert=config.samples_per_insert,
        min_replay_size=min_replay_size,
        num_sgd_steps_per_step=config.num_sgd_steps_per_step,
        counter=counter,
    )



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reverb utils.

Contains functions manipulating reverb tables and samples.
"""

from acme import types
import jax
import numpy as np
import reverb
from reverb import item_selectors
from reverb import rate_limiters
from reverb import reverb_types
import tree


def make_replay_table_from_info(
    table_info: reverb_types.TableInfo) -> reverb.Table:
  """Build a replay table out of its specs in a TableInfo.

  Args:
    table_info: A TableInfo containing the Table specs.

  Returns:
    A reverb replay table matching the info specs.
  """
  sampler = _make_selector_from_key_distribution_options(
      table_info.sampler_options)
  remover = _make_selector_from_key_distribution_options(
      table_info.remover_options)
  rate_limiter = _make_rate_limiter_from_rate_limiter_info(
      table_info.rate_limiter_info)
  return reverb.Table(
      name=table_info.name,
      sampler=sampler,
      remover=remover,
      max_size=table_info.max_size,
      rate_limiter=rate_limiter,
      max_times_sampled=table_info.max_times_sampled,
      signature=table_info.signature)


def _make_selector_from_key_distribution_options(
    options) -> reverb_types.SelectorType:
  """Returns a Selector from its KeyDistributionOptions description."""
  one_of = options.WhichOneof('distribution')
  if one_of == 'fifo':
    return item_selectors.Fifo()
  if one_of == 'uniform':
    return item_selectors.Uniform()
  if one_of == 'prioritized':
    return item_selectors.Prioritized(options.prioritized.priority_exponent)
  if one_of == 'heap':
    if options.heap.min_heap:
      return item_selectors.MinHeap()
    return item_selectors.MaxHeap()
  if one_of == 'lifo':
    return item_selectors.Lifo()
  raise ValueError(f'Unknown distribution field: {one_of}')


def _make_rate_limiter_from_rate_limiter_info(
    info) -> rate_limiters.RateLimiter:
  return rate_limiters.SampleToInsertRatio(
      samples_per_insert=info.samples_per_insert,
      min_size_to_sample=info.min_size_to_sample,
      error_buffer=(info.min_diff, info.max_diff))


def replay_sample_to_sars_transition(
    sample: reverb.ReplaySample,
    is_sequence: bool,
    strip_last_transition: bool = False,
    flatten_batch: bool = False) -> types.Transition:
  """Converts the replay sample to a types.Transition.

  NB: If is_sequence is True then the last next_observation of each sequence is
  rubbish. Don't train on it.

  Args:
    sample: The replay sample
    is_sequence: If False we expect the sample data to match the
      types.Transition already. Otherwise we expect a batch of sequences of
      steps.
    strip_last_transition: If True and is_sequence, the last transition will be
      stripped as its next_observation field is incorrect.
    flatten_batch: If True and is_sequence, the two batch dimensions will be
      flatten to one.

  Returns:
    A types.Transition built from the sample data.
    If is_sequence and strip_last_transition are both True, the output will be
    smaller than the output as the last transition of every sequence will have
    been removed.
  """
  if not is_sequence:
    return types.Transition(*sample.data)
  # Note that the last next_observation is invalid.
  steps = sample.data
  def roll(observation):
    return np.roll(observation, shift=-1, axis=1)
  transitions = types.Transition(
      observation=steps.observation,
      action=steps.action,
      reward=steps.reward,
      discount=steps.discount,
      next_observation=tree.map_structure(roll, steps.observation))
  if strip_last_transition:
    # We remove the last transition as its next_observation field is incorrect.
    # It has been obtained by rolling the observation field, such that
    # transitions.next_observations[:, -1] is transitions.observations[:, 0]
    transitions = jax.tree_map(lambda x: x[:, :-1, ...], transitions)
  if flatten_batch:
    # Merge the 2 leading batch dimensions into 1.
    transitions = jax.tree_map(lambda x: np.reshape(x, (-1,) + x.shape[2:]),
                               transitions)
  return transitions



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for image logger."""

import os

from absl.testing import absltest
from acme.testing import test_utils
from acme.utils.loggers import image
import numpy as np
from PIL import Image


class ImageTest(test_utils.TestCase):

  def test_save_load_identity(self):
    directory = self.get_tempdir()
    logger = image.ImageLogger(directory, label='foo')
    array = (np.random.rand(10, 10) * 255).astype(np.uint8)
    logger.write({'img': array})

    with open(f'{directory}/foo/img_000000.png', mode='rb') as f:
      out = np.asarray(Image.open(f))
    np.testing.assert_array_equal(array, out)

  def test_indexing(self):
    directory = self.get_tempdir()
    logger = image.ImageLogger(directory, label='foo')
    zeros = np.zeros(shape=(3, 3), dtype=np.uint8)
    logger.write({'img': zeros, 'other_img': zeros + 1})
    logger.write({'img': zeros - 1})
    logger.write({'other_img': zeros + 1})
    logger.write({'other_img': zeros + 2})

    fnames = sorted(os.listdir(f'{directory}/foo'))
    expected = [
        'img_000000.png',
        'img_000001.png',
        'other_img_000000.png',
        'other_img_000001.png',
        'other_img_000002.png',
    ]
    self.assertEqual(fnames, expected)


if __name__ == '__main__':
  absltest.main()



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the AIL discriminator losses."""

from absl.testing import absltest
from acme import types
from acme.agents.jax.ail import losses
from acme.jax import networks as networks_lib
import jax
from jax import test_util as jtu
import jax.numpy as jnp
import tree


class AilLossTest(jtu.JaxTestCase):

  def test_gradient_penalty(self):

    def dummy_discriminator(
        transition: types.Transition) -> networks_lib.Logits:
      return transition.observation + jnp.square(transition.action)

    zero_transition = types.Transition(0., 0., 0., 0., 0.)
    zero_transition = tree.map_structure(lambda x: jnp.expand_dims(x, axis=0),
                                         zero_transition)
    self.assertEqual(
        losses._compute_gradient_penalty(zero_transition, dummy_discriminator,
                                         0.), 1**2 + 0**2)

    one_transition = types.Transition(1., 1., 0., 0., 0.)
    one_transition = tree.map_structure(lambda x: jnp.expand_dims(x, axis=0),
                                        one_transition)
    self.assertEqual(
        losses._compute_gradient_penalty(one_transition, dummy_discriminator,
                                         0.), 1**2 + 2**2)

  def test_pugail(self):

    def dummy_discriminator(
        state: losses.State,
        transition: types.Transition) -> losses.DiscriminatorOutput:
      return transition.observation, state

    zero_transition = types.Transition(.1, 0., 0., 0., 0.)
    zero_transition = tree.map_structure(lambda x: jnp.expand_dims(x, axis=0),
                                         zero_transition)

    one_transition = types.Transition(1., 0., 0., 0., 0.)
    one_transition = tree.map_structure(lambda x: jnp.expand_dims(x, axis=0),
                                        one_transition)

    prior = .7
    loss_fn = losses.pugail_loss(
        positive_class_prior=prior, entropy_coefficient=0.)
    loss, _ = loss_fn(dummy_discriminator, {}, one_transition,
                      zero_transition, ())

    d_one = jax.nn.sigmoid(dummy_discriminator({}, one_transition)[0])
    d_zero = jax.nn.sigmoid(dummy_discriminator({}, zero_transition)[0])
    expected_loss = -prior * jnp.log(
        d_one) + -jnp.log(1. - d_zero) - prior * -jnp.log(1 - d_one)

    self.assertAlmostEqual(loss, expected_loss, places=6)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AIL config."""
import dataclasses
from typing import Optional

import optax


@dataclasses.dataclass
class AILConfig:
  """Configuration options for AIL.

  Attributes:
    direct_rl_batch_size: Batch size of a direct rl algorithm (measured in
      transitions).
    is_sequence_based: If True, a direct rl algorithm is using SequenceAdder
      data format. Otherwise the learner assumes that the direct rl algorithm is
      using NStepTransitionAdder.
    share_iterator: If True, AIL will use the same iterator for the
      discriminator network training as the direct rl algorithm.
    num_sgd_steps_per_step: Only used if 'share_iterator' is False. Denotes how
      many gradient updates perform per one learner step.
    discriminator_batch_size:  Batch size for training the discriminator.
    policy_variable_name: The name of the policy variable to retrieve direct_rl
      policy parameters.
    discriminator_optimizer: Optimizer for the discriminator. If not specified
      it is set to Adam with learning rate of 1e-5.
    replay_table_name: The name of the reverb replay table to use.
    prefetch_size: How many batches to prefetch
    discount: Discount to use for TD updates
    min_replay_size: Minimal size of replay buffer
    max_replay_size: Maximal size of replay buffer
    policy_to_expert_data_ratio: If not None, the direct RL learner will receive
      expert transitions in the given proportions.
      policy_to_expert_data_ratio + 1 must divide the direct RL batch size.
  """
  direct_rl_batch_size: int
  is_sequence_based: bool = False
  share_iterator: bool = True
  num_sgd_steps_per_step: int = 1
  discriminator_batch_size: int = 256
  policy_variable_name: Optional[str] = None
  discriminator_optimizer: Optional[optax.GradientTransformation] = None
  replay_table_name: str = 'ail_table'
  prefetch_size: int = 4
  discount: float = 0.99
  min_replay_size: int = 1000
  max_replay_size: int = int(1e6)
  policy_to_expert_data_ratio: Optional[int] = None

  def __post_init__(self):
    assert self.direct_rl_batch_size % self.discriminator_batch_size == 0


def get_per_learner_step_batch_size(config: AILConfig) -> int:
  """Returns how many transitions should be sampled per direct learner step."""
  # If the iterators are tied, the discriminator learning batch size has to
  # match the direct RL one.
  if config.share_iterator:
    assert (config.direct_rl_batch_size % config.discriminator_batch_size) == 0
    return config.direct_rl_batch_size
  # Otherwise each iteration of the discriminator will sample a batch which will
  # be split in num_sgd_steps_per_step batches, each of size
  # discriminator_batch_size.
  return config.discriminator_batch_size * config.num_sgd_steps_per_step



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility function for building and launching launchpad programs."""

import atexit
import functools
import inspect
import os
import sys
import time
from typing import Any, Callable, Optional

from absl import flags
from absl import logging
from acme.utils import counting
from acme.utils import signals

FLAGS = flags.FLAGS


def partial_kwargs(function: Callable[..., Any],
                   **kwargs: Any) -> Callable[..., Any]:
  """Return a partial function application by overriding default keywords.

  This function is equivalent to `functools.partial(function, **kwargs)` but
  will raise a `ValueError` when called if either the given keyword arguments
  are not defined by `function` or if they do not have defaults.

  This is useful as a way to define a factory function with default parameters
  and then to override them in a safe way.

  Args:
    function: the base function before partial application.
    **kwargs: keyword argument overrides.

  Returns:
    A function.
  """
  # Try to get the argspec of our function which we'll use to get which keywords
  # have defaults.
  argspec = inspect.getfullargspec(function)

  # Figure out which keywords have defaults.
  if argspec.defaults is None:
    defaults = []
  else:
    defaults = argspec.args[-len(argspec.defaults):]

  # Find any keys not given as defaults by the function.
  unknown_kwargs = set(kwargs.keys()).difference(defaults)

  # Raise an error
  if unknown_kwargs:
    error_string = 'Cannot override unknown or non-default kwargs: {}'
    raise ValueError(error_string.format(', '.join(unknown_kwargs)))

  return functools.partial(function, **kwargs)


class StepsLimiter:
  """Process that terminates an experiment when `max_steps` is reached."""

  def __init__(self,
               counter: counting.Counter,
               max_steps: int,
               steps_key: str = 'actor_steps'):
    self._counter = counter
    self._max_steps = max_steps
    self._steps_key = steps_key

  def run(self):
    """Run steps limiter to terminate an experiment when max_steps is reached.
    """

    logging.info('StepsLimiter: Starting with max_steps = %d (%s)',
                 self._max_steps, self._steps_key)
    with signals.runtime_terminator():
      while True:
        # Update the counts.
        counts = self._counter.get_counts()
        num_steps = counts.get(self._steps_key, 0)

        logging.info('StepsLimiter: Reached %d recorded steps', num_steps)

        if num_steps > self._max_steps:
          logging.info('StepsLimiter: Max steps of %d was reached, terminating',
                       self._max_steps)
          # Avoid importing Launchpad until it is actually used.
          import launchpad as lp  # pylint: disable=g-import-not-at-top
          lp.stop()

        # Don't spam the counter.
        for _ in range(10):
          # Do not sleep for a long period of time to avoid LaunchPad program
          # termination hangs (time.sleep is not interruptible).
          time.sleep(1)


# Resources for each individual instance of the program.
def make_xm_docker_resources(program,
                             requirements: Optional[str] = None):
  """Returns Docker XManager resources for each program's node.

  For each node of the Launchpad's program appropriate hardware requirements are
  specified (CPU, memory...), while the list of PyPi packages specified in
  the requirements file will be installed inside the Docker images.

  Args:
    program: program for which to construct Docker XManager resources.
    requirements: file containing additional requirements to use.
      If not specified, default Acme dependencies are used instead.
  """
  if (FLAGS.lp_launch_type != 'vertex_ai' and
      FLAGS.lp_launch_type != 'local_docker'):
    # Avoid importing 'xmanager' for local runs.
    return None

  # Avoid importing Launchpad until it is actually used.
  import launchpad as lp  # pylint: disable=g-import-not-at-top
  # Reference lp.DockerConfig to force lazy import of xmanager by Launchpad and
  # then import it. It is done this way to avoid heavy imports by default.
  lp.DockerConfig  # pylint: disable=pointless-statement
  from xmanager import xm  # pylint: disable=g-import-not-at-top

  # Get number of each type of node.
  num_nodes = {k: len(v) for k, v in program.groups.items()}

  xm_resources = {}

  acme_location = os.path.dirname(os.path.dirname(__file__))
  if not requirements:
    # Acme requirements are located in the Acme directory (when installed
    # with pip), or need to be extracted from setup.py when using Acme codebase
    # from GitHub without PyPi installation.
    requirements = os.path.join(acme_location, 'requirements.txt')
    if not os.path.isfile(requirements):
      # Try to generate requirements.txt from setup.py
      setup = os.path.join(os.path.dirname(acme_location), 'setup.py')
      if os.path.isfile(setup):
        # Generate requirements.txt file using setup.py.
        import importlib.util  # pylint: disable=g-import-not-at-top
        spec = importlib.util.spec_from_file_location('setup', setup)
        setup = importlib.util.module_from_spec(spec)
        try:
          spec.loader.exec_module(setup)  # pytype: disable=attribute-error
        except SystemExit:
          pass
        atexit.register(os.remove, requirements)
        setup.generate_requirements_file(requirements)

  # Extend PYTHONPATH with paths used by the launcher.
  python_path = []
  for path in sys.path:
    if path.startswith(acme_location) and acme_location != path:
      python_path.append(path[len(acme_location):])

  if 'replay' in num_nodes:
    replay_cpu = 6 + num_nodes.get('actor', 0) * 0.01
    replay_cpu = min(40, replay_cpu)

    xm_resources['replay'] = lp.DockerConfig(
        acme_location,
        requirements,
        hw_requirements=xm.JobRequirements(cpu=replay_cpu, ram=10 * xm.GiB),
        python_path=python_path)

  if 'evaluator' in num_nodes:
    xm_resources['evaluator'] = lp.DockerConfig(
        acme_location,
        requirements,
        hw_requirements=xm.JobRequirements(cpu=2, ram=4 * xm.GiB),
        python_path=python_path)

  if 'actor' in num_nodes:
    xm_resources['actor'] = lp.DockerConfig(
        acme_location,
        requirements,
        hw_requirements=xm.JobRequirements(cpu=2, ram=4 * xm.GiB),
        python_path=python_path)

  if 'learner' in num_nodes:
    learner_cpu = 6 + num_nodes.get('actor', 0) * 0.01
    learner_cpu = min(40, learner_cpu)
    xm_resources['learner'] = lp.DockerConfig(
        acme_location,
        requirements,
        hw_requirements=xm.JobRequirements(
            cpu=learner_cpu, ram=6 * xm.GiB, P100=1),
        python_path=python_path)

  if 'environment_loop' in num_nodes:
    xm_resources['environment_loop'] = lp.DockerConfig(
        acme_location,
        requirements,
        hw_requirements=xm.JobRequirements(
            cpu=6, ram=6 * xm.GiB, P100=1),
        python_path=python_path)

  if 'counter' in num_nodes:
    xm_resources['counter'] = lp.DockerConfig(
        acme_location,
        requirements,
        hw_requirements=xm.JobRequirements(cpu=3, ram=4 * xm.GiB),
        python_path=python_path)

  if 'cacher' in num_nodes:
    xm_resources['cacher'] = lp.DockerConfig(
        acme_location,
        requirements,
        hw_requirements=xm.JobRequirements(cpu=3, ram=6 * xm.GiB),
        python_path=python_path)

  return xm_resources



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the BC agent."""

from absl.testing import absltest
from absl.testing import parameterized
from acme import specs
from acme import types
from acme.agents.jax import bc
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.testing import fakes
import chex
import haiku as hk
import jax
import jax.numpy as jnp
from jax.scipy import special
import numpy as np
import optax


def make_networks(
    spec: specs.EnvironmentSpec,
    discrete_actions: bool = False) -> networks_lib.FeedForwardNetwork:
  """Creates networks used by the agent."""

  if discrete_actions:
    final_layer_size = spec.actions.num_values
  else:
    final_layer_size = np.prod(spec.actions.shape, dtype=int)

  def _actor_fn(obs, is_training=False, key=None):
    # is_training and key allows to defined train/test dependant modules
    # like dropout.
    del is_training
    del key
    if discrete_actions:
      network = hk.nets.MLP([64, 64, final_layer_size])
    else:
      network = hk.Sequential([
          networks_lib.LayerNormMLP([64, 64], activate_final=True),
          networks_lib.NormalTanhDistribution(final_layer_size),
      ])
    return network(obs)

  policy = hk.without_apply_rng(hk.transform(_actor_fn))

  # Create dummy observations and actions to create network parameters.
  dummy_obs = utils.zeros_like(spec.observations)
  dummy_obs = utils.add_batch_dim(dummy_obs)
  network = networks_lib.FeedForwardNetwork(
      lambda key: policy.init(key, dummy_obs), policy.apply)
  return network


class BCTest(parameterized.TestCase):

  @parameterized.parameters(
      ('logp',),
      ('mse',),
      ('peerbc',)
      )
  def test_continuous_actions(self, loss_name):
    with chex.fake_pmap_and_jit():
      num_sgd_steps_per_step = 1
      num_steps = 5

      # Create a fake environment to test with.
      environment = fakes.ContinuousEnvironment(
          episode_length=10, bounded=True, action_dim=6)

      spec = specs.make_environment_spec(environment)
      dataset_demonstration = fakes.transition_dataset(environment)
      dataset_demonstration = dataset_demonstration.map(
          lambda sample: types.Transition(*sample.data))
      dataset_demonstration = dataset_demonstration.batch(8).as_numpy_iterator()

      # Construct the agent.
      network = make_networks(spec)

      if loss_name == 'logp':
        loss_fn = bc.logp(
            logp_fn=lambda dist_params, actions: dist_params.log_prob(actions))
      elif loss_name == 'mse':
        loss_fn = bc.mse(
            sample_fn=lambda dist_params, key: dist_params.sample(seed=key))
      elif loss_name == 'peerbc':
        base_loss_fn = bc.logp(
            logp_fn=lambda dist_params, actions: dist_params.log_prob(actions))
        loss_fn = bc.peerbc(base_loss_fn, zeta=0.1)
      else:
        raise ValueError

      learner = bc.BCLearner(
          network=network,
          random_key=jax.random.PRNGKey(0),
          loss_fn=loss_fn,
          optimizer=optax.adam(0.01),
          demonstrations=dataset_demonstration,
          num_sgd_steps_per_step=num_sgd_steps_per_step)

      # Train the agent
      for _ in range(num_steps):
        learner.step()

  @parameterized.parameters(
      ('logp',),
      ('rcal',))
  def test_discrete_actions(self, loss_name):
    with chex.fake_pmap_and_jit():

      num_sgd_steps_per_step = 1
      num_steps = 5

      # Create a fake environment to test with.
      environment = fakes.DiscreteEnvironment(
          num_actions=10, num_observations=100, obs_shape=(10,),
          obs_dtype=np.float32)

      spec = specs.make_environment_spec(environment)
      dataset_demonstration = fakes.transition_dataset(environment)
      dataset_demonstration = dataset_demonstration.map(
          lambda sample: types.Transition(*sample.data))
      dataset_demonstration = dataset_demonstration.batch(8).as_numpy_iterator()

      # Construct the agent.
      network = make_networks(spec, discrete_actions=True)

      def logp_fn(logits, actions):
        max_logits = jnp.max(logits, axis=-1, keepdims=True)
        logits = logits - max_logits
        logits_actions = jnp.sum(
            jax.nn.one_hot(actions, spec.actions.num_values) * logits, axis=-1)

        log_prob = logits_actions - special.logsumexp(logits, axis=-1)
        return log_prob

      if loss_name == 'logp':
        loss_fn = bc.logp(logp_fn=logp_fn)

      elif loss_name == 'rcal':
        base_loss_fn = bc.logp(logp_fn=logp_fn)
        loss_fn = bc.rcal(base_loss_fn, discount=0.99, alpha=0.1)

      else:
        raise ValueError

      learner = bc.BCLearner(
          network=network,
          random_key=jax.random.PRNGKey(0),
          loss_fn=loss_fn,
          optimizer=optax.adam(0.01),
          demonstrations=dataset_demonstration,
          num_sgd_steps_per_step=num_sgd_steps_per_step)

      # Train the agent
      for _ in range(num_steps):
        learner.step()


if __name__ == '__main__':
  absltest.main()



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""NoOp Starts wrapper to allow stochastic initial state for deterministic Python environments."""

from typing import Optional

from acme import types
from acme.wrappers import base
import dm_env
import numpy as np


class NoopStartsWrapper(base.EnvironmentWrapper):
  """Implements random noop starts to episodes.

  This introduces randomness into an otherwise deterministic environment.

  Note that the base environment must support a no-op action and the value
  of this action must be known and provided to this wrapper.
  """

  def __init__(self,
               environment: dm_env.Environment,
               noop_action: types.NestedArray = 0,
               noop_max: int = 30,
               seed: Optional[int] = None):
    """Initializes a `NoopStartsWrapper` wrapper.

    Args:
      environment: An environment conforming to the dm_env.Environment
        interface.
      noop_action: The noop action used to step the environment for random
        initialisation.
      noop_max: The maximal number of noop actions at the start of an episode.
      seed: The random seed used to sample the number of noops.
    """
    if noop_max < 0:
      raise ValueError(
          'Maximal number of no-ops after reset cannot be negative. '
          f'Received noop_max={noop_max}')

    super().__init__(environment)
    self.np_random = np.random.RandomState(seed)
    self._noop_max = noop_max
    self._noop_action = noop_action

  def reset(self) -> dm_env.TimeStep:
    """Resets environment and provides the first timestep."""
    noops = self.np_random.randint(self._noop_max + 1)
    timestep = self.environment.reset()
    for _ in range(noops):
      timestep = self.environment.step(self._noop_action)
      if timestep.last():
        timestep = self.environment.reset()

    return timestep._replace(step_type=dm_env.StepType.FIRST)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for running statistics utilities."""

import functools
import math
from typing import NamedTuple

from absl.testing import absltest
from acme import specs
from acme.jax import running_statistics
import jax
from jax import test_util as jtu
from jax.config import config as jax_config
import jax.numpy as jnp
import tree

update_and_validate = functools.partial(
    running_statistics.update, validate_shapes=True)


class TestNestedSpec(NamedTuple):
  # Note: the fields are intentionally in reverse order to test ordering.
  a: specs.Array
  b: specs.Array


class RunningStatisticsTest(jtu.JaxTestCase):

  def setUp(self):
    super().setUp()
    jax_config.update('jax_enable_x64', False)

  def test_normalize(self):
    state = running_statistics.init_state(specs.Array((5,), jnp.float32))

    x = jnp.arange(200, dtype=jnp.float32).reshape(20, 2, 5)
    x1, x2, x3, x4 = jnp.split(x, 4, axis=0)

    state = update_and_validate(state, x1)
    state = update_and_validate(state, x2)
    state = update_and_validate(state, x3)
    state = update_and_validate(state, x4)
    normalized = running_statistics.normalize(x, state)

    mean = jnp.mean(normalized)
    std = jnp.std(normalized)
    self.assertAllClose(mean, jnp.zeros_like(mean))
    self.assertAllClose(std, jnp.ones_like(std))

  def test_init_normalize(self):
    state = running_statistics.init_state(specs.Array((5,), jnp.float32))

    x = jnp.arange(200, dtype=jnp.float32).reshape(20, 2, 5)
    normalized = running_statistics.normalize(x, state)

    self.assertAllClose(normalized, x)

  def test_one_batch_dim(self):
    state = running_statistics.init_state(specs.Array((5,), jnp.float32))

    x = jnp.arange(10, dtype=jnp.float32).reshape(2, 5)

    state = update_and_validate(state, x)
    normalized = running_statistics.normalize(x, state)

    mean = jnp.mean(normalized, axis=0)
    std = jnp.std(normalized, axis=0)
    self.assertAllClose(mean, jnp.zeros_like(mean))
    self.assertAllClose(std, jnp.ones_like(std))

  def test_clip(self):
    state = running_statistics.init_state(specs.Array((), jnp.float32))

    x = jnp.arange(5, dtype=jnp.float32)

    state = update_and_validate(state, x)
    normalized = running_statistics.normalize(x, state, max_abs_value=1.0)

    mean = jnp.mean(normalized)
    std = jnp.std(normalized)
    self.assertAllClose(mean, jnp.zeros_like(mean))
    self.assertAllClose(std, jnp.ones_like(std) * math.sqrt(0.6))

  def test_nested_normalize(self):
    state = running_statistics.init_state({
        'a': specs.Array((5,), jnp.float32),
        'b': specs.Array((2,), jnp.float32)
    })

    x1 = {
        'a': jnp.arange(20, dtype=jnp.float32).reshape(2, 2, 5),
        'b': jnp.arange(8, dtype=jnp.float32).reshape(2, 2, 2)
    }
    x2 = {
        'a': jnp.arange(20, dtype=jnp.float32).reshape(2, 2, 5) + 20,
        'b': jnp.arange(8, dtype=jnp.float32).reshape(2, 2, 2) + 8
    }
    x3 = {
        'a': jnp.arange(40, dtype=jnp.float32).reshape(4, 2, 5),
        'b': jnp.arange(16, dtype=jnp.float32).reshape(4, 2, 2)
    }

    state = update_and_validate(state, x1)
    state = update_and_validate(state, x2)
    state = update_and_validate(state, x3)
    normalized = running_statistics.normalize(x3, state)

    mean = tree.map_structure(lambda x: jnp.mean(x, axis=(0, 1)), normalized)
    std = tree.map_structure(lambda x: jnp.std(x, axis=(0, 1)), normalized)
    tree.map_structure(lambda x: self.assertAllClose(x, jnp.zeros_like(x)),
                       mean)
    tree.map_structure(lambda x: self.assertAllClose(x, jnp.ones_like(x)), std)

  def test_validation(self):
    state = running_statistics.init_state(specs.Array((1, 2, 3), jnp.float32))

    x = jnp.arange(12, dtype=jnp.float32).reshape(2, 2, 3)
    with self.assertRaises(AssertionError):
      update_and_validate(state, x)

    x = jnp.arange(3, dtype=jnp.float32).reshape(1, 1, 3)
    with self.assertRaises(AssertionError):
      update_and_validate(state, x)

  def test_int_not_normalized(self):
    state = running_statistics.init_state(specs.Array((), jnp.int32))

    x = jnp.arange(5, dtype=jnp.int32)

    state = update_and_validate(state, x)
    normalized = running_statistics.normalize(x, state)

    self.assertArraysEqual(normalized, x)

  def test_pmap_update_nested(self):
    local_device_count = jax.local_device_count()
    assert local_device_count > 1
    state = running_statistics.init_state({
        'a': specs.Array((5,), jnp.float32),
        'b': specs.Array((2,), jnp.float32)
    })

    x = {
        'a': (jnp.arange(15 * local_device_count,
                         dtype=jnp.float32)).reshape(local_device_count, 3, 5),
        'b': (jnp.arange(6 * local_device_count,
                         dtype=jnp.float32)).reshape(local_device_count, 3, 2),
    }

    devices = jax.local_devices()
    state = jax.device_put_replicated(state, devices)
    pmap_axis_name = 'i'
    state = jax.pmap(
        functools.partial(update_and_validate, pmap_axis_name=pmap_axis_name),
        pmap_axis_name)(state, x)
    state = jax.pmap(
        functools.partial(update_and_validate, pmap_axis_name=pmap_axis_name),
        pmap_axis_name)(state, x)
    normalized = jax.pmap(running_statistics.normalize)(x, state)

    mean = tree.map_structure(lambda x: jnp.mean(x, axis=(0, 1)), normalized)
    std = tree.map_structure(lambda x: jnp.std(x, axis=(0, 1)), normalized)
    tree.map_structure(lambda x: self.assertAllClose(x, jnp.zeros_like(x)),
                       mean)
    tree.map_structure(lambda x: self.assertAllClose(x, jnp.ones_like(x)), std)

  def test_different_structure_normalize(self):
    spec = TestNestedSpec(
        a=specs.Array((5,), jnp.float32), b=specs.Array((2,), jnp.float32))
    state = running_statistics.init_state(spec)

    x = {
        'a': jnp.arange(20, dtype=jnp.float32).reshape(2, 2, 5),
        'b': jnp.arange(8, dtype=jnp.float32).reshape(2, 2, 2)
    }

    with self.assertRaises(TypeError):
      state = update_and_validate(state, x)

  def test_weights(self):
    state = running_statistics.init_state(specs.Array((), jnp.float32))

    x = jnp.arange(5, dtype=jnp.float32)
    x_weights = jnp.ones_like(x)
    y = 2 * x + 5
    y_weights = 2 * x_weights
    z = jnp.concatenate([x, y])
    weights = jnp.concatenate([x_weights, y_weights])

    state = update_and_validate(state, z, weights=weights)

    self.assertEqual(state.mean, (jnp.mean(x) + 2 * jnp.mean(y)) / 3)
    big_z = jnp.concatenate([x, y, y])
    normalized = running_statistics.normalize(big_z, state)
    self.assertAlmostEqual(jnp.mean(normalized), 0., places=6)
    self.assertAlmostEqual(jnp.std(normalized), 1., places=6)

  def test_normalize_config(self):
    x = jnp.arange(200, dtype=jnp.float32).reshape(20, 2, 5)
    x_split = jnp.split(x, 5, axis=0)

    y = jnp.arange(160, dtype=jnp.float32).reshape(20, 2, 4)
    y_split = jnp.split(y, 5, axis=0)

    z = {'a': x, 'b': y}

    z_split = [{'a': xx, 'b': yy} for xx, yy in zip(x_split, y_split)]

    update = jax.jit(running_statistics.update, static_argnames=('config',))

    config = running_statistics.NestStatisticsConfig((('a',),))
    state = running_statistics.init_state({
        'a': specs.Array((5,), jnp.float32),
        'b': specs.Array((4,), jnp.float32)
    })
    # Test initialization from the first element.
    state = update(state, z_split[0], config=config)
    state = update(state, z_split[1], config=config)
    state = update(state, z_split[2], config=config)
    state = update(state, z_split[3], config=config)
    state = update(state, z_split[4], config=config)

    normalize = jax.jit(running_statistics.normalize)
    normalized = normalize(z, state)

    for key in normalized:
      mean = jnp.mean(normalized[key], axis=(0, 1))
      std = jnp.std(normalized[key], axis=(0, 1))
      if key == 'a':
        self.assertAllClose(
            mean,
            jnp.zeros_like(mean),
            err_msg=f'key:{key} mean:{mean} normalized:{normalized[key]}')
        self.assertAllClose(
            std,
            jnp.ones_like(std),
            err_msg=f'key:{key} std:{std} normalized:{normalized[key]}')
      else:
        assert key == 'b'
        self.assertArraysEqual(
            normalized[key],
            z[key],
            err_msg=f'z:{z[key]} normalized:{normalized[key]}')

  def test_clip_config(self):
    x = jnp.arange(10, dtype=jnp.float32) - 5
    y = jnp.arange(8, dtype=jnp.float32) - 4

    z = {'x': x, 'y': y}

    max_abs_x = 2
    config = running_statistics.NestClippingConfig(((('x',), max_abs_x),))

    clipped_z = running_statistics.clip(z, config)

    clipped_x = jnp.clip(a=x, a_min=-max_abs_x, a_max=max_abs_x)
    self.assertArraysEqual(clipped_z['x'], clipped_x)

    self.assertArraysEqual(clipped_z['y'], z['y'])

  def test_denormalize(self):
    state = running_statistics.init_state(specs.Array((5,), jnp.float32))

    x = jnp.arange(100, dtype=jnp.float32).reshape(10, 2, 5)
    x1, x2 = jnp.split(x, 2, axis=0)

    state = update_and_validate(state, x1)
    state = update_and_validate(state, x2)
    normalized = running_statistics.normalize(x, state)

    mean = jnp.mean(normalized)
    std = jnp.std(normalized)
    self.assertAllClose(mean, jnp.zeros_like(mean))
    self.assertAllClose(std, jnp.ones_like(std))

    denormalized = running_statistics.denormalize(normalized, state)
    self.assertAllClose(denormalized, x)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MPO learner implementation."""

import time
from typing import List, Optional

import acme
from acme import types
from acme.tf import losses
from acme.tf import savers as tf2_savers
from acme.tf import utils as tf2_utils
from acme.utils import counting
from acme.utils import loggers
import numpy as np
import sonnet as snt
import tensorflow as tf
import trfl


class MPOLearner(acme.Learner):
  """MPO learner."""

  def __init__(
      self,
      policy_network: snt.Module,
      critic_network: snt.Module,
      target_policy_network: snt.Module,
      target_critic_network: snt.Module,
      discount: float,
      num_samples: int,
      target_policy_update_period: int,
      target_critic_update_period: int,
      dataset: tf.data.Dataset,
      observation_network: types.TensorTransformation = tf.identity,
      target_observation_network: types.TensorTransformation = tf.identity,
      policy_loss_module: Optional[snt.Module] = None,
      policy_optimizer: Optional[snt.Optimizer] = None,
      critic_optimizer: Optional[snt.Optimizer] = None,
      dual_optimizer: Optional[snt.Optimizer] = None,
      clipping: bool = True,
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
      checkpoint: bool = True,
      save_directory: str = '~/acme',
  ):

    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger('learner')
    self._discount = discount
    self._num_samples = num_samples
    self._clipping = clipping

    # Necessary to track when to update target networks.
    self._num_steps = tf.Variable(0, dtype=tf.int32)
    self._target_policy_update_period = target_policy_update_period
    self._target_critic_update_period = target_critic_update_period

    # Batch dataset and create iterator.
    # TODO(b/155086959): Fix type stubs and remove.
    self._iterator = iter(dataset)  # pytype: disable=wrong-arg-types

    # Store online and target networks.
    self._policy_network = policy_network
    self._critic_network = critic_network
    self._target_policy_network = target_policy_network
    self._target_critic_network = target_critic_network

    # Make sure observation networks are snt.Module's so they have variables.
    self._observation_network = tf2_utils.to_sonnet_module(observation_network)
    self._target_observation_network = tf2_utils.to_sonnet_module(
        target_observation_network)

    self._policy_loss_module = policy_loss_module or losses.MPO(
        epsilon=1e-1,
        epsilon_penalty=1e-3,
        epsilon_mean=1e-3,
        epsilon_stddev=1e-6,
        init_log_temperature=1.,
        init_log_alpha_mean=1.,
        init_log_alpha_stddev=10.)

    # Create the optimizers.
    self._critic_optimizer = critic_optimizer or snt.optimizers.Adam(1e-4)
    self._policy_optimizer = policy_optimizer or snt.optimizers.Adam(1e-4)
    self._dual_optimizer = dual_optimizer or snt.optimizers.Adam(1e-2)

    # Expose the variables.
    policy_network_to_expose = snt.Sequential(
        [self._target_observation_network, self._target_policy_network])
    self._variables = {
        'critic': self._target_critic_network.variables,
        'policy': policy_network_to_expose.variables,
    }

    # Create a checkpointer and snapshotter object.
    self._checkpointer = None
    self._snapshotter = None

    if checkpoint:
      self._checkpointer = tf2_savers.Checkpointer(
          directory=save_directory,
          subdirectory='mpo_learner',
          objects_to_save={
              'counter': self._counter,
              'policy': self._policy_network,
              'critic': self._critic_network,
              'observation_network': self._observation_network,
              'target_policy': self._target_policy_network,
              'target_critic': self._target_critic_network,
              'target_observation_network': self._target_observation_network,
              'policy_optimizer': self._policy_optimizer,
              'critic_optimizer': self._critic_optimizer,
              'dual_optimizer': self._dual_optimizer,
              'policy_loss_module': self._policy_loss_module,
              'num_steps': self._num_steps,
          })

      self._snapshotter = tf2_savers.Snapshotter(
          directory=save_directory,
          objects_to_save={
              'policy':
                  snt.Sequential([
                      self._target_observation_network,
                      self._target_policy_network
                  ]),
          })

    # Do not record timestamps until after the first learning step is done.
    # This is to avoid including the time it takes for actors to come online and
    # fill the replay buffer.
    self._timestamp = None

  @tf.function
  def _step(self) -> types.Nest:
    # Update target network.
    online_policy_variables = self._policy_network.variables
    target_policy_variables = self._target_policy_network.variables
    online_critic_variables = (
        *self._observation_network.variables,
        *self._critic_network.variables,
    )
    target_critic_variables = (
        *self._target_observation_network.variables,
        *self._target_critic_network.variables,
    )

    # Make online policy -> target policy network update ops.
    if tf.math.mod(self._num_steps, self._target_policy_update_period) == 0:
      for src, dest in zip(online_policy_variables, target_policy_variables):
        dest.assign(src)
    # Make online critic -> target critic network update ops.
    if tf.math.mod(self._num_steps, self._target_critic_update_period) == 0:
      for src, dest in zip(online_critic_variables, target_critic_variables):
        dest.assign(src)

    # Increment number of learner steps for periodic update bookkeeping.
    self._num_steps.assign_add(1)

    # Get next batch of data.
    inputs = next(self._iterator)

    # Get data from replay (dropping extras if any). Note there is no
    # extra data here because we do not insert any into Reverb.
    transitions: types.Transition = inputs.data

    # Cast the additional discount to match the environment discount dtype.
    discount = tf.cast(self._discount, dtype=transitions.discount.dtype)

    with tf.GradientTape(persistent=True) as tape:
      # Maybe transform the observation before feeding into policy and critic.
      # Transforming the observations this way at the start of the learning
      # step effectively means that the policy and critic share observation
      # network weights.
      o_tm1 = self._observation_network(transitions.observation)
      # This stop_gradient prevents gradients to propagate into the target
      # observation network. In addition, since the online policy network is
      # evaluated at o_t, this also means the policy loss does not influence
      # the observation network training.
      o_t = tf.stop_gradient(
          self._target_observation_network(transitions.next_observation))

      # Get action distributions from policy networks.
      online_action_distribution = self._policy_network(o_t)
      target_action_distribution = self._target_policy_network(o_t)

      # Get sampled actions to evaluate policy; of size [N, B, ...].
      sampled_actions = target_action_distribution.sample(self._num_samples)
      tiled_o_t = tf2_utils.tile_tensor(o_t, self._num_samples)  # [N, B, ...]

      # Compute the target critic's Q-value of the sampled actions in state o_t.
      sampled_q_t = self._target_critic_network(
          # Merge batch dimensions; to shape [N*B, ...].
          snt.merge_leading_dims(tiled_o_t, num_dims=2),
          snt.merge_leading_dims(sampled_actions, num_dims=2))

      # Reshape Q-value samples back to original batch dimensions and average
      # them to compute the TD-learning bootstrap target.
      sampled_q_t = tf.reshape(sampled_q_t, (self._num_samples, -1))  # [N, B]
      q_t = tf.reduce_mean(sampled_q_t, axis=0)  # [B]

      # Compute online critic value of a_tm1 in state o_tm1.
      q_tm1 = self._critic_network(o_tm1, transitions.action)  # [B, 1]
      q_tm1 = tf.squeeze(q_tm1, axis=-1)  # [B]; necessary for trfl.td_learning.

      # Critic loss.
      critic_loss = trfl.td_learning(q_tm1, transitions.reward,
                                     discount * transitions.discount, q_t).loss
      critic_loss = tf.reduce_mean(critic_loss)

      # Actor learning.
      policy_loss, policy_stats = self._policy_loss_module(
          online_action_distribution=online_action_distribution,
          target_action_distribution=target_action_distribution,
          actions=sampled_actions,
          q_values=sampled_q_t)

    # For clarity, explicitly define which variables are trained by which loss.
    critic_trainable_variables = (
        # In this agent, the critic loss trains the observation network.
        self._observation_network.trainable_variables +
        self._critic_network.trainable_variables)
    policy_trainable_variables = self._policy_network.trainable_variables
    # The following are the MPO dual variables, stored in the loss module.
    dual_trainable_variables = self._policy_loss_module.trainable_variables

    # Compute gradients.
    critic_gradients = tape.gradient(critic_loss, critic_trainable_variables)
    policy_gradients, dual_gradients = tape.gradient(
        policy_loss, (policy_trainable_variables, dual_trainable_variables))

    # Delete the tape manually because of the persistent=True flag.
    del tape

    # Maybe clip gradients.
    if self._clipping:
      policy_gradients = tuple(tf.clip_by_global_norm(policy_gradients, 40.)[0])
      critic_gradients = tuple(tf.clip_by_global_norm(critic_gradients, 40.)[0])

    # Apply gradients.
    self._critic_optimizer.apply(critic_gradients, critic_trainable_variables)
    self._policy_optimizer.apply(policy_gradients, policy_trainable_variables)
    self._dual_optimizer.apply(dual_gradients, dual_trainable_variables)

    # Losses to track.
    fetches = {
        'critic_loss': critic_loss,
        'policy_loss': policy_loss,
    }
    fetches.update(policy_stats)  # Log MPO stats.

    return fetches

  def step(self):
    # Run the learning step.
    fetches = self._step()

    # Compute elapsed time.
    timestamp = time.time()
    elapsed_time = timestamp - self._timestamp if self._timestamp else 0
    self._timestamp = timestamp

    # Update our counts and record it.
    counts = self._counter.increment(steps=1, walltime=elapsed_time)
    fetches.update(counts)

    # Checkpoint and attempt to write the logs.
    if self._checkpointer is not None:
      self._checkpointer.save()
    if self._snapshotter is not None:
      self._snapshotter.save()
    self._logger.write(fetches)

  def get_variables(self, names: List[str]) -> List[List[np.ndarray]]:
    return [tf2_utils.to_numpy(self._variables[name]) for name in names]



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logger which self closes on exit if not closed yet."""

import weakref

from acme.utils.loggers import base


class AutoCloseLogger(base.Logger):
  """Logger which auto closes itself on exit if not already closed."""

  def __init__(self, logger: base.Logger):
    self._logger = logger
    # The finalizer "logger.close" is invoked in one of the following scenario:
    # 1) the current logger is GC
    # 2) from the python doc, when the program exits, each remaining live
    #    finalizer is called.
    # Note that in the normal flow, where "close" is explicitly called,
    # the finalizer is marked as dead using the detach function so that
    # the underlying logger is not closed twice (once explicitly and once
    # implicitly when the object is GC or when the program exits).
    self._finalizer = weakref.finalize(self, logger.close)

  def write(self, values: base.LoggingData):
    self._logger.write(values)

  def close(self):
    if self._finalizer.detach():
      self._logger.close()
    self._logger = None



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Learner for the IMPALA actor-critic agent."""

import time
from typing import Dict, Iterator, List, NamedTuple, Optional, Sequence, Tuple

from absl import logging
import acme
from acme.agents.jax.impala import networks as impala_networks
from acme.jax import losses
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.utils import counting
from acme.utils import loggers
import jax
import jax.numpy as jnp
import numpy as np
import optax
import reverb

_PMAP_AXIS_NAME = 'data'


class TrainingState(NamedTuple):
  """Training state consists of network parameters and optimiser state."""
  params: networks_lib.Params
  opt_state: optax.OptState


class IMPALALearner(acme.Learner):
  """Learner for an importanced-weighted advantage actor-critic."""

  def __init__(
      self,
      networks: impala_networks.IMPALANetworks,
      iterator: Iterator[reverb.ReplaySample],
      optimizer: optax.GradientTransformation,
      random_key: networks_lib.PRNGKey,
      discount: float = 0.99,
      entropy_cost: float = 0.,
      baseline_cost: float = 1.,
      max_abs_reward: float = np.inf,
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
      devices: Optional[Sequence[jax.xla.Device]] = None,
      prefetch_size: int = 2,
      num_prefetch_threads: Optional[int] = None,
  ):
    local_devices = jax.local_devices()
    process_id = jax.process_index()
    logging.info('Learner process id: %s. Devices passed: %s', process_id,
                 devices)
    logging.info('Learner process id: %s. Local devices from JAX API: %s',
                 process_id, local_devices)
    self._devices = devices or local_devices
    self._local_devices = [d for d in self._devices if d in local_devices]

    loss_fn = losses.impala_loss(
        networks.unroll_fn,
        discount=discount,
        max_abs_reward=max_abs_reward,
        baseline_cost=baseline_cost,
        entropy_cost=entropy_cost)

    @jax.jit
    def sgd_step(
        state: TrainingState, sample: reverb.ReplaySample
    ) -> Tuple[TrainingState, Dict[str, jnp.ndarray]]:
      """Computes an SGD step, returning new state and metrics for logging."""

      # Compute gradients.
      grad_fn = jax.value_and_grad(loss_fn)
      loss_value, gradients = grad_fn(state.params, sample)

      # Average gradients over pmap replicas before optimizer update.
      gradients = jax.lax.pmean(gradients, _PMAP_AXIS_NAME)

      # Apply updates.
      updates, new_opt_state = optimizer.update(gradients, state.opt_state)
      new_params = optax.apply_updates(state.params, updates)

      metrics = {
          'loss': loss_value,
          'param_norm': optax.global_norm(new_params),
          'param_updates_norm': optax.global_norm(updates),
      }

      new_state = TrainingState(params=new_params, opt_state=new_opt_state)

      return new_state, metrics

    def make_initial_state(key: jnp.ndarray) -> TrainingState:
      """Initialises the training state (parameters and optimiser state)."""
      key, key_initial_state = jax.random.split(key)
      # Note: parameters do not depend on the batch size, so initial_state below
      # does not need a batch dimension.
      params = networks.initial_state_init_fn(key_initial_state)
      # TODO(jferret): as it stands, we do not yet support
      # training the initial state params.
      initial_state = networks.initial_state_fn(params)

      initial_params = networks.unroll_init_fn(key, initial_state)
      initial_opt_state = optimizer.init(initial_params)
      return TrainingState(
          params=initial_params, opt_state=initial_opt_state)

    # Initialise training state (parameters and optimiser state).
    state = make_initial_state(random_key)
    self._state = utils.replicate_in_all_devices(state, self._local_devices)

    if num_prefetch_threads is None:
      num_prefetch_threads = len(self._local_devices)
    self._prefetched_iterator = utils.sharded_prefetch(
        iterator,
        buffer_size=prefetch_size,
        devices=self._local_devices,
        num_threads=num_prefetch_threads,
    )

    self._sgd_step = jax.pmap(
        sgd_step, axis_name=_PMAP_AXIS_NAME, devices=self._devices)

    # Set up logging/counting.
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger('learner')

  def step(self):
    """Does a step of SGD and logs the results."""
    samples = next(self._prefetched_iterator)

    # Do a batch of SGD.
    start = time.time()
    self._state, results = self._sgd_step(self._state, samples)

    # Take results from first replica.
    results = utils.get_from_first_device(results)

    # Update our counts and record it.
    counts = self._counter.increment(steps=1, time_elapsed=time.time() - start)

    # Snapshot and attempt to write logs.
    self._logger.write({**results, **counts})

  def get_variables(self, names: Sequence[str]) -> List[networks_lib.Params]:
    # Return first replica of parameters.
    return [utils.get_from_first_device(self._state.params, as_numpy=False)]

  def save(self) -> TrainingState:
    # Serialize only the first replica of parameters and optimizer state.
    return jax.tree_map(utils.get_from_first_device, self._state)

  def restore(self, state: TrainingState):
    self._state = utils.replicate_in_all_devices(state, self._local_devices)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common environment wrapper classes."""

from acme.wrappers.action_repeat import ActionRepeatWrapper
from acme.wrappers.atari_wrapper import AtariWrapper
from acme.wrappers.base import EnvironmentWrapper
from acme.wrappers.base import wrap_all
from acme.wrappers.canonical_spec import CanonicalSpecWrapper
from acme.wrappers.concatenate_observations import ConcatObservationWrapper
from acme.wrappers.frame_stacking import FrameStackingWrapper
from acme.wrappers.gym_wrapper import GymAtariAdapter
from acme.wrappers.gym_wrapper import GymWrapper
from acme.wrappers.noop_starts import NoopStartsWrapper
from acme.wrappers.observation_action_reward import ObservationActionRewardWrapper
from acme.wrappers.single_precision import SinglePrecisionWrapper
from acme.wrappers.step_limit import StepLimitWrapper

try:
  # pylint: disable=g-import-not-at-top
  from acme.wrappers.open_spiel_wrapper import OpenSpielWrapper
except ImportError:
  pass



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Losses and projection operators relevant to distributional RL."""

from acme.tf import networks
import tensorflow as tf


def categorical(q_tm1: networks.DiscreteValuedDistribution, r_t: tf.Tensor,
                d_t: tf.Tensor,
                q_t: networks.DiscreteValuedDistribution) -> tf.Tensor:
  """Implements the Categorical Distributional TD(0)-learning loss."""

  z_t = tf.reshape(r_t, (-1, 1)) + tf.reshape(d_t, (-1, 1)) * q_t.values
  p_t = tf.nn.softmax(q_t.logits)

  # Performs L2 projection.
  target = tf.stop_gradient(l2_project(z_t, p_t, q_t.values))

  # Calculates loss.
  loss = tf.nn.softmax_cross_entropy_with_logits(
      logits=q_tm1.logits, labels=target)

  return loss


# Use an old version of the l2 projection which is probably slower on CPU
# but will run on GPUs.
def l2_project(  # pylint: disable=invalid-name
    Zp: tf.Tensor,
    P: tf.Tensor,
    Zq: tf.Tensor,
) -> tf.Tensor:
  """Project distribution (Zp, P) onto support Zq under the L2-metric over CDFs.

  This projection works for any support Zq.
  Let Kq be len(Zq) and Kp be len(Zp).

  Args:
    Zp: (batch_size, Kp) Support of distribution P
    P:  (batch_size, Kp) Probability values for P(Zp[i])
    Zq: (Kp,) Support to project onto

  Returns:
    L2 projection of (Zp, P) onto Zq.
  """

  # Asserts that Zq has no leading dimension of size 1.
  if Zq.get_shape().ndims > 1:
    Zq = tf.squeeze(Zq, axis=0)

  # Extracts vmin and vmax and construct helper tensors from Zq.
  vmin, vmax = Zq[0], Zq[-1]
  d_pos = tf.concat([Zq, vmin[None]], 0)[1:]
  d_neg = tf.concat([vmax[None], Zq], 0)[:-1]

  # Clips Zp to be in new support range (vmin, vmax).
  clipped_zp = tf.clip_by_value(Zp, vmin, vmax)[:, None, :]
  clipped_zq = Zq[None, :, None]

  # Gets the distance between atom values in support.
  d_pos = (d_pos - Zq)[None, :, None]  # Zq[i+1] - Zq[i]
  d_neg = (Zq - d_neg)[None, :, None]  # Zq[i] - Zq[i-1]

  delta_qp = clipped_zp - clipped_zq  # Zp[j] - Zq[i]

  d_sign = tf.cast(delta_qp >= 0., dtype=P.dtype)
  delta_hat = (d_sign * delta_qp / d_pos) - ((1. - d_sign) * delta_qp / d_neg)
  P = P[:, None, :]
  return tf.reduce_sum(tf.clip_by_value(1. - delta_hat, 0., 1.) * P, 2)


def multiaxis_categorical(  # pylint: disable=invalid-name
    q_tm1: networks.DiscreteValuedDistribution,
    r_t: tf.Tensor,
    d_t: tf.Tensor,
    q_t: networks.DiscreteValuedDistribution) -> tf.Tensor:
  """Implements a multi-axis categorical distributional TD(0)-learning loss.

  All arguments may have a leading batch axis, but q_tm1.logits, and one of
  r_t or d_t *must* have a leading batch axis.

  Args:
    q_tm1: Previous timestep's value distribution.
    r_t: Reward.
    d_t: Discount.
    q_t: Current timestep's value distribution.

  Returns:
    Cross-entropy Bellman loss between q_tm1 and q_t + r_t * d_t.
    Shape: (B, *E), where
      B is the batch size.
      E is the broadcasted shape of r_t, d_t, and q_t.values[:-1].
  """
  tf.assert_equal(tf.rank(r_t), tf.rank(d_t))

  # Append a singleton axis corresponding to the axis that indexes the atoms in
  # q_t.values.
  r_t = r_t[..., None]  # shape: (B, *R, 1)
  d_t = d_t[..., None]  # shape: (B, *D, 1)

  z_t = r_t + d_t * q_t.values  # shape: (B, *E, N)

  p_t = tf.nn.softmax(q_t.logits)

  # Performs L2 projection.
  target = tf.stop_gradient(multiaxis_l2_project(z_t, p_t, q_t.values))

  # Calculates loss.
  loss = tf.nn.softmax_cross_entropy_with_logits(
      logits=q_tm1.logits, labels=target)

  return loss


# A modification of l2_project that allows multi-axis support arguments.
def multiaxis_l2_project(  # pylint: disable=invalid-name
    Zp: tf.Tensor,
    P: tf.Tensor,
    Zq: tf.Tensor,
) -> tf.Tensor:
  """Project distribution (Zp, P) onto support Zq under the L2-metric over CDFs.

  Let source support Zp's shape be described as (B, *C, M), where:
    B is the batch size.
    C contains the sizes of any axes in between the first and last axes.
    M is the number of atoms in the support.

  Let destination support Zq's shape be described as (*D, N), where:
    D contains the sizes of any axes before the last axis.
    N is the number of atoms in the support.

  Shapes C and D must have the same number of dimensions, and must be
  broadcastable with each other.

  Args:
    Zp: Support of source distribution. Shape: (B, *C, M).
    P:  Probability values of source distribution p(Zp[i]). Shape: (B, *C, M).
    Zq: Support to project P onto. Shape: (*D, N).

  Returns:
    The L2 projection of P from support Zp to support Zq.
    Shape: (B, *E, N), where E is the broadcast-merged shape of C and D.
  """

  tf.assert_equal(tf.shape(Zp), tf.shape(P))

  # Shapes C, D, and E as defined in the docstring above.
  shape_c = tf.shape(Zp)[1:-1]  # drop the batch and atom axes
  shape_d = tf.shape(Zq)[:-1]  # drop the atom axis
  shape_e = tf.broadcast_dynamic_shape(shape_c, shape_d)

  # If Zq has fewer inner axes than the broadcasted output shape, insert some
  # size-1 axes to broadcast.
  ndim_c = tf.size(shape_c)
  ndim_e = tf.size(shape_e)
  Zp = tf.reshape(
      Zp,
      tf.concat([tf.shape(Zp)[:1],  # B
                 tf.ones(tf.math.maximum(ndim_e - ndim_c, 0), dtype=tf.int32),
                 shape_c,  # C
                 tf.shape(Zp)[-1:]],  # M
                axis=0))
  P = tf.reshape(P, tf.shape(Zp))

  # Broadcast Zp, P, and Zq's common axes to the same shape: E.
  #
  # Normally it'd be sufficient to ensure that these args have the same number
  # of axes, then let the arithmetic operators broadcast as necessary. Instead,
  # we need to explicitly broadcast them here, because there's a call to
  # tf.clip_by_value(t, vmin, vmax) below, which doesn't allow t's dimensions
  # to be expanded to match vmin and vmax.

  # Shape: (B, *E, M)
  Zp = tf.broadcast_to(
      Zp,
      tf.concat([tf.shape(Zp)[:1],  # B
                 shape_e,  # E
                 tf.shape(Zp)[-1:]],  # M
                axis=0))

  # Shape: (B, *E, M)
  P = tf.broadcast_to(P, tf.shape(Zp))

  # Shape: (*E, N)
  Zq = tf.broadcast_to(Zq, tf.concat([shape_e, tf.shape(Zq)[-1:]], axis=0))

  # Extracts vmin and vmax and construct helper tensors from Zq.
  # These have shape shape_q, except the last axis has size 1.
  # Shape: (*E, 1)
  vmin, vmax = Zq[..., :1], Zq[..., -1:]

  # The distances between neighboring atom values in the target support.
  # Shape: (*E, N)
  d_pos = tf.roll(Zq, shift=-1, axis=-1) - Zq  # d_pos[i] := Zq[i+1] - Zq[i]
  d_neg = Zq - tf.roll(Zq, shift=1, axis=-1)   # d_neg[i] := Zq[i] - Zq[i-1]

  # Clips Zp to be in new support range (vmin, vmax).
  # Shape: (B, *E, 1, M)
  clipped_zp = tf.clip_by_value(Zp, vmin, vmax)[..., None, :]

  # Shape: (1, *E, N, 1)
  clipped_zq = Zq[None, ..., :, None]

  # Shape: (B, *E, N, M)
  delta_qp = clipped_zp - clipped_zq  # Zp[j] - Zq[i]

  # Shape: (B, *E, N, M)
  d_sign = tf.cast(delta_qp >= 0., dtype=P.dtype)

  # Insert singleton axes to d_pos and d_neg to maintain the same shape as
  # clipped_zq.
  # Shape: (1, *E, N, 1)
  d_pos = d_pos[None, ..., :, None]
  d_neg = d_neg[None, ..., :, None]

  # Shape: (B, *E, N, M)
  delta_hat = (d_sign * delta_qp / d_pos) - ((1. - d_sign) * delta_qp / d_neg)

  # Shape: (B, *E, 1, M)
  P = P[..., None, :]

  # Shape: (B, *E, N)
  return tf.reduce_sum(tf.clip_by_value(1. - delta_hat, 0., 1.) * P, axis=-1)
